/*
 * uart3.h
 *
 *  Created on: Sep 17, 2021
 *      Author: thomm
 */

#ifndef UART3_H_
#define UART3_H_

void uart3_init(void);
void uart3_setup(uint32_t ui32Baud, uint32_t ui32SrcClock);

int uart3_write(const char *pcBuf, uint32_t ui32Len);
int uart3_rx_str(char *pcBuf, uint32_t ui32Len);
int uart3_rx_buf(unsigned char *pcBuf, uint32_t ui32Len);

unsigned char uart3_rx_byte(void);
void uart3_tx_byte(unsigned char c);

int uart3_rx_bytes_avail(void);
int uart3_tx_bufr_bytes_free(void);
int uart3_rx_bufr_peek(unsigned char ucChar);
void uart3_rx_bufr_flush(void);
void uart3_tx_bufr_flush(bool bDiscard);

void uart3_echo_set(bool bEnable);

void uart3_isr(void);           // plug this into interrupt table in tm4c123gh6pge_startup_ccs.c




#endif /* UART3_H_ */
